const path = require("path")
const net = require("net");
const fs = require("fs");
const os = require("os");
const crypto = require("crypto");
const {execSync, exec} = require("child_process");
const {mkdir, access, constants} = require("node:fs/promises")
const {app} = require('electron')
const iconv = require("iconv-lite");

const tar = require("tar");
const http = require("http");

async function prepareFolder() {
    /**
     * appData 每个用户的应用程序数据目录，默认情况下指向：
     %APPDATA% Windows 中
     $XDG_CONFIG_HOME or ~/.config Linux 中
     ~/Library/Application Support macOS 中
     */

    const topDir = app.getPath("userData")
    const dbDir = path.join(topDir, "pgsql")
    const redisDir = path.join(topDir, "redis")
    const tmpDir = path.join(topDir, "tmp")
    const logDir = path.join(topDir, "logs")

    await ensureDir(dbDir)
    await ensureDir(redisDir)
    await ensureDir(tmpDir)
    await ensureDir(logDir)

    return {
        dbDir, tmpDir, logDir, redisDir
    }
}

/**
 * 获取默认存储文件夹 - 缩略图、上传的文件夹
 */
async function getDefaultRootFolder() {
    const folderName = "mt-photos-files"
    let list = []
    try {
        list = await getDistList()
    } catch (e) {
    }
    if (list.length === 0 || !(list instanceof Array)) {
        //fallback
        try {
            list = await getDistListInPS()
        } catch (e) {
        }
    }

    if (list instanceof Array) {
        list = list.filter(d => d.driveType === 3).sort((a, b) => b.freeSpace - a.freeSpace)
        if (list.length > 0) {
            return path.join(list[0].caption, folderName)
        }
    }

    const homeDir = app.getPath("home")
    return path.join(homeDir, folderName)
}

function getDbDir() {
    const topDir = app.getPath("userData")
    return path.join(topDir, "pgsql")
}

function getLogDir() {
    const topDir = app.getPath("userData")
    return path.join(topDir, "logs")
}

function getTmpDir() {
    const topDir = app.getPath("userData")
    return path.join(topDir, "tmp", "__mt_photos_temp")
}

function getUserDataDir() {
    return app.getPath("userData")
}

async function isPathExist(filePath) {
    try {
        await access(filePath, constants.R_OK | constants.W_OK);
        return true
    } catch {
    }
    return false
}

async function ensureDir(filePath) {
    const isExist = await isPathExist(filePath)
    if (!isExist) {
        await mkdir(filePath, {recursive: true});
    }
}

/**
 * 获取硬盘列表
 * @returns {*[]}
 */
async function getDistList() {
    // node-disk-info  https://github.com/cristiammercado/node-disk-info
    const cmd_win = 'wmic logicaldisk get Caption,FreeSpace,Size,DriveType  /format:list'
    const cmd_osx = 'df -P | awk \'NR > 1\''
    const drives = [];
    let buffer = await execAsync(cmd_win, false);
    const cp = chcp();
    let encoding = '';
    switch (cp) {
        case '65000':
            encoding = 'UTF-7';
            break;
        case '65001':
            encoding = 'UTF-8';
            break;
        default:
            if (/^-?[\d.]+(?:e-?\d+)?$/.test(cp)) {
                encoding = 'cp' + cp;
            } else {
                encoding = cp;
            }
    }
    buffer = iconv.encode(iconv.decode(buffer, encoding), 'UTF-8');
    const lines = buffer.toString().split('\r\r\n');
    let newDiskIteration = false;

    let caption = '';
    let driveType = '';
    let freeSpace = 0;
    let size = 0;

    /**
     *
     driveType列表
     0    Unknown
     1    No Root Directory
     2    Removable Disk
     3    Local Disk
     4    Network Drive
     5    Compact Disc
     6    RAM Disk
     */

    lines.forEach((value) => {

        if (value !== '') {

            const tokens = value.split('=');
            const section = tokens[0];
            const data = tokens[1];

            switch (section) {
                case 'Caption':
                    caption = data;
                    newDiskIteration = true;
                    break;
                case 'DriveType':
                    driveType = Number(data);
                    break;
                case 'FreeSpace':
                    freeSpace = isNaN(parseFloat(data)) ? 0 : +data;
                    break;
                case 'Size':
                    size = isNaN(parseFloat(data)) ? 0 : +data;
                    break;
            }

        } else {

            if (newDiskIteration) {

                drives.push({
                    driveType,
                    size,
                    freeSpace,
                    caption
                });
                newDiskIteration = false;
                caption = '';
                driveType = 0;
                freeSpace = 0;
                size = 0;
            }

        }

    });
    return drives;
}

function getDistListInPS() {
    return new Promise((resolve, reject) => {
        const command = 'powershell -Command "Get-CimInstance -ClassName Win32_LogicalDisk | Select-Object DeviceID, VolumeName, Size, FreeSpace, DriveType"';
        exec(command, (error, stdout, stderr) => {
            if (error) {
                resolve([])
                return;
            }

            if (stderr) {
                resolve([])
                return;
            }

            const lines = stdout.toString().split('\r\n');
            let drives = []
            let newDiskIteration = false;
            let caption = '';
            let driveType = 0;
            let freeSpace = 0;
            let size = 0;
            /**
             *
             driveType列表
             0    Unknown
             1    No Root Directory
             2    Removable Disk
             3    Local Disk
             4    Network Drive
             5    Compact Disc
             6    RAM Disk
             */

            lines.forEach((value) => {

                if (value !== '') {
                    const tokens = value.split(':');
                    let section = tokens[0];
                    let data = tokens[1];
                    if (!section || !section.trim) {
                        return
                    }
                    section = section.trim()
                    data = data.trim()
                    switch (section) {
                        case 'DeviceID':
                            caption = data + ":"; //这边用:分割
                            newDiskIteration = true;
                            break;
                        case 'DriveType':
                            driveType = Number(data);
                            break;
                        case 'FreeSpace':
                            freeSpace = isNaN(parseFloat(data)) ? 0 : +data;
                            break;
                        case 'Size':
                            size = isNaN(parseFloat(data)) ? 0 : +data;
                            break;
                    }

                } else {
                    if (newDiskIteration) {
                        drives.push({
                            driveType,
                            size,
                            freeSpace,
                            caption
                        });
                        newDiskIteration = false;
                        caption = '';
                        driveType = 0;
                        freeSpace = 0;
                        size = 0;
                    }

                }

            });
            resolve(drives);
        });
    })
}

function chcp() {
    return execSync('chcp').toString().split(':')[1].trim();
}

async function autoSetPort() {
    const pgPort = await getRandomPort(28032);
    const redisPort = await getRandomPort(28179);
    return {
        pgPort,
        redisPort
    }
}

async function getRandomPort(defaultPort) {
    let isUsed = await checkPort('127.0.0.1', defaultPort);
    if (!isUsed) {
        return defaultPort;
    }
    for (let i = 0; i < 99; i++) {
        let _port = defaultPort + i;
        isUsed = await checkPort('127.0.0.1', _port);
        if (!isUsed) {
            return _port;
        }
    }
}

function checkPort(server, port) {
    return new Promise(async (resolve, reject) => {
        let client = net.createConnection({host: server, port: port, timeout: 100}, () => {
            client.end();
            resolve(true);
        });
        client.on('error', e => {
            resolve(false);
        });
    });
}

async function waitPortConnect(port, maxTry = 100) {
    let i = 0
    while (i < maxTry) {
        i++
        let isConnect = await checkPort('127.0.0.1', port);
        if (isConnect) {
            return true
        }
        await wait(100)
    }
    return false
}

function wait(ms = 500) {
    return new Promise((resolve, reject) => {
        setTimeout(resolve, ms)
    })
}

function readFile(filePath) {
    return new Promise((resolve, reject) => {
        fs.readFile(filePath, 'utf8', (err, data) => {
            if (err) {
                reject(err);
                return;
            }
            resolve(data)
        })
    })
}

function getIpList() {
    let interfaces = os.networkInterfaces();
    let ip = [];
    for (let devName in interfaces) {
        let iface = interfaces[devName];
        for (let i = 0; i < iface.length; i++) {
            let alias = iface[i];
            if (alias.family === 'IPv4' && alias.address !== '127.0.0.1' && !alias.internal) {
                if (/^172\./.test(alias.address)) {
                    continue
                }
                ip.push(alias.address);
            }
        }
    }
    return ip
}

async function upgrade(msg, dist) {
    const {source} = msg.data
    if (!source || await isPathExist(source) === false) {
        return {msg: "sourceError"}
    }
    await unTar(source, dist)
    return {}
}

function unTar(source, dist) {
    return new Promise((resolve, reject) => {
        const rs = fs.createReadStream(source).pipe(
            tar.x({
                cwd: dist
            })
        )
        rs.on('end', function () {
            resolve()
        });
        rs.on('error', function (e) {
            reject(e)
        });
    })
}

function md5(str) {
    return crypto.createHash('md5').update(str).digest("hex");
}

async function getSystemInfo() {
    const ffmpeg = await execAsync('ffmpeg -version', true)
    const ImageMagick = await execAsync('magick -version', true)
    return {
        cpu: os.cpus()[0].model + " X" + os.cpus().length,
        mem: (os.totalmem() / 1024 / 1024 / 1024).toFixed(0) + "GB",
        platform: os.platform() + " " + os.arch(),
        version: os.version() + " " + os.release(),
        uptime: (os.uptime() / 60 / 60).toFixed(2) + "h",
        ImageMagick,
        ffmpeg,
    }
}

function execAsync(command, resolveErr) {
    return new Promise((resolve, reject) => {
        exec(command, {encoding: "utf8"}, (e, res) => {
            if (e) {
                if (resolveErr === true) {
                    resolve(e)
                } else {
                    reject(e)
                }
                return
            }
            resolve(res)
        })
    })
}

function tryDeleteFile(filePath) {
    return new Promise((resolve, reject) => {
        try {
            fs.unlink(filePath, () => {
                resolve()
            });
        } catch (e) {
        }
    })
}

function isWindowsRootDirectory(p) {
    const normalizedPath = path.normalize(p);
    const regex = /^[A-Za-z]:[\\\/]?$/;
    return regex.test(normalizedPath);
}


function checkExistBackupFile({interval, file_num, nowDate, existsFiles}) {
    let needDeleteFiles = []
    let dayNum = interval * file_num
    const now = nowDate.getTime()
    let latestFileDate = 0
    for (let i = 0; i < existsFiles.length; i++) {
        const {name, path} = existsFiles[i];
        const fileDate = getAutoBackupDate(name)
        if (fileDate.isBackupFile) {
            if ((now - fileDate.dateTime + 1000 * 60 * 60) > 1000 * 60 * 60 * 24 * dayNum) {
                needDeleteFiles.push(path);
            } else {
                if (fileDate.dateTime > latestFileDate) {
                    latestFileDate = fileDate.dateTime
                }
            }
        }
    }

    if ((now - latestFileDate + 1000 * 60 * 60) < 1000 * 60 * 60 * 24 * interval) {
        return {needAdd: false, needDeleteFiles}
    }
    return {needAdd: true, needDeleteFiles}
}

function getAutoBackupDate(filename) {
    if (filename.includes("mt_photos") && filename.includes("auto_db_dump")) {
        const regex = /(\d{4}-\d{2}-\d{2})_(\d{2})_(\d{2})_(\d{2})/;
        const match = filename.match(regex);

        if (match) {
            const dateString = match[1] + 'T' + match[2] + ':' + match[3] + ':' + match[4];
            const date = new Date(dateString);
            return {isBackupFile: true, dateTime: date.getTime()}
        }
    }
    return {isBackupFile: false}
}

function readdir(dir) {
    return new Promise((resolve, reject) => {
        fs.readdir(dir, (err, list) => {
            if (err) {
                resolve([]);
            } else {
                resolve(list);
            }
        });
    });
}

function isDirAsync(path) {
    return new Promise((resolve, reject) => {
        fs.stat(path, (err, stat) => {
            if (err) {
                resolve(false);
            }
            resolve(stat && stat.isDirectory());
        });
    });
}

async function _find(dir) {
    let list = [];
    try {
        let arr = await readdir(dir);
        for (let i = 0; i < arr.length; i++) {
            const item = arr[i];
            const fullPath = path.join(dir, item);
            if (await isDirAsync(fullPath)) {
            } else {
                list.push({
                    name: item,
                    path: fullPath
                });
            }
        }
    } catch (e) {

    }
    return list;
}

function joinPath(...args) {
    return path.join.apply(null, args);
}

function getFileData(path){
    return new Promise((resolve, reject) => {
        fs.readFile(path,{encoding:"utf8"},(err,result)=>{
            if(err){
                return resolve("")
            }
            resolve(result)
        })
    });
}

function writeFileAsync(filePath, data) {
    return new Promise((resolve, reject) => {
        fs.writeFile(filePath, data, (err) => {
            if (err) {
                return reject(err);
            }
            return resolve(filePath);
        });
    });
}


function getServiceStatus(url) {
    return new Promise((resolve, reject) => {
        let timeout
        const req = http.get(url, (res) => {
            let data = '';
            res.on('data', (chunk) => {
                data += chunk;
            });
            res.on('end', () => {
                clearTimeout(timeout)
                try {
                    const jsonResponse = JSON.parse(data);
                    if (jsonResponse.dbStatus && jsonResponse.dbStatus.pgEnable === false) {
                        resolve(false)
                    } else {
                        resolve(true)
                    }
                } catch (error) {
                    resolve(false)
                }
            });
        }).on('error', (err) => {
            clearTimeout(timeout)
            resolve(false)
        });
        timeout = setTimeout(() => {
            req.destroy();
        }, 1000);
    })
}

function checkServerStatus({log,restartServer,appStatus}) {
    let checker = setInterval(() => {
        // 调用方法检查服务状态
        if(appStatus.serverStatus !== "running"){
            return
        }
        getServiceStatus(`http://127.0.0.1:${appStatus.serverPort}/api-info`).then((res) => {
            if (res === false) {
                let now = new Date()
                if (log) {
                    log.info("restart server "+ now.toString())
                }
                if (restartServer) {
                    restartServer()
                }
                clearInterval(checker)
                return
            }
        }).catch((e) => {
        })
    }, [5 * 60 * 1000])
    return checker
}

module.exports = {
    prepareFolder,
    getDbDir,
    getLogDir,
    getTmpDir,
    getUserDataDir,
    isPathExist,
    getDistList,
    getDefaultRootFolder,
    ensureDir,
    autoSetPort,
    getRandomPort,
    waitPortConnect,
    readFile,
    getIpList,
    upgrade,
    md5,
    getSystemInfo,
    wait,
    tryDeleteFile,
    isWindowsRootDirectory,
    _find, checkExistBackupFile,
    joinPath,
    writeFileAsync,checkServerStatus
}
